from __future__ import division

import numpy as np
from numpy.testing import assert_equal
from pyoperators.utils.testing import assert_same
from pysimulators import (
    create_fitsheader, LayoutGridSquares, Quantity, SceneGrid)
from pysimulators.geometry import create_grid_squares

itypes = np.int32, np.int64
ftypes = np.float32, np.float64

def test_spatial_integration():
    #  -------------
    #  |12|13|14|15|
    #  |--+--+--+--+
    #  | 8| 9|10|11|
    #  |--+--+--+--+
    #  | 4| 5| 6| 7|
    #  |--+--+--+--+
    #  | 0| 1| 2| 3|
    #  -------------
    #  The plane center has world coordinates (0, 0)

    plane_shape = (4, 4)
    pixsize = 1  # um
    header = create_fitsheader(plane_shape, ctype=['X---CAR', 'Y---CAR'],
                               cdelt=[pixsize, pixsize],
                               crval=(0, 0), cunit=['um', 'um'])
    plane = SceneGrid.fromfits(header)

    def func(center, ncolmax, itype, ftype, expected):
        detectors = create_grid_squares((2, 2), Quantity(pixsize, 'um'),
                                        center=center)
        proj = plane.get_integration_operator(
            plane.topixel(detectors), ncolmax=ncolmax, dtype=ftype,
            dtype_index=itype)
        x = np.arange(len(plane)).reshape(plane_shape)
        y = proj(x)

        assert_equal(proj.matrix.dtype, ftype)
        assert_equal(proj.matrix.data.index.dtype, itype)
        assert_equal(proj.shapein, plane.shape)
        assert_equal(proj.shapeout, detectors.shape[:-2])
        assert_equal(proj.matrix.shape, (4, len(plane)))
        expected_min_ncolmax = 1 if center == (0, 0) else 4
        assert_equal(proj.matrix.data.shape, (4, max(expected_min_ncolmax,
                                                     ncolmax)))
        assert_equal(proj.min_ncolmax, expected_min_ncolmax)
        assert_same(y, expected)

    centers = [(0, 0), (-0.5, 0.5)]
    ncolmaxs = (0, 1, 4, 5)
    expecteds = [[[9, 10],
                  [5, 6]],
                 [[0.25*(8+9+12+13), 0.25*(9+10+13+14)],
                  [0.25*(4+5+8+9),   0.25*(5+6+9+10)]]]
    for center, expected in zip(centers, expecteds):
        for ncolmax in ncolmaxs:
            for itype in itypes:
                for ftype in ftypes:
                    yield func, center, ncolmax, itype, ftype, expected


def test_spatial_integration2():
    shape_grid = (16, 16)
    spacings = 0.5, 1, 2
    angles = 0, 30
    filling_factor = 0.8

    def func(spacing, angle):
        origin = shape_grid * np.array(spacing) * np.sqrt(2) / 2
        grid = LayoutGridSquares(shape_grid, spacing, origin=origin,
                                 angle=angle, filling_factor=filling_factor)
        shape_plane = np.ceil(shape_grid * np.array(spacing) * np.sqrt(2))
        scene = SceneGrid(shape_plane)
        proj = scene.get_integration_operator(grid.vertex)
        assert not proj.outside
        assert_same(np.sum(proj.matrix.data.value),
                    len(grid) * spacing**2 * filling_factor, rtol=100)

    for spacing in spacings:
        for angle in angles:
            yield func, spacing, angle



from __future__ import absolute_import, division, print_function
from pyoperators import (IdentityOperator, To1dOperator, asoperator)
from pyoperators.utils import strenum, product
from .core import PackedTable
from .. import _flib as flib
from ..operators import ProjectionOperator
from ..sparse import FSRMatrix
from ..wcsutils import fitsheader2shape, WCSToPixelOperator
import numpy as np

__all__ = ['Scene', 'SceneGrid']


class Scene(PackedTable):
    """
    Class to handle plane pixelisation and 1-d indexing.

    """
    def __init__(self, shape, topixel=None, ndim=None, **keywords):
        """
    Parameters
    ----------
    shape : tuple of integers
        The surface shape.
    ndim : int, optional
        The number of splittable (indexable) dimensions. It is the actual
        number of dimensions of the layout. It can be lower than that
        specified by the layout shape, in which case the extra dimensions
        are instructed not to be split.
    topixel : Operator, optional
        World-to-pixel coordinate transform.

        """
        PackedTable.__init__(self, shape, ndim=ndim)
        if topixel is None:
            topixel = IdentityOperator(self.shape[:self.ndim])
        else:
            topixel = asoperator(topixel)
        self.topixel = topixel
        self.toworld = topixel.I
        for k, v in keywords.items():
            setattr(self, k, v)


class SceneGrid(Scene):
    """
    Class for 2-dimensional scenes.

    """
    def __init__(self, shape, topixel=None, to1d=None, origin='upper',
                 startswith1=False, **keywords):
        """
        Parameters
        ----------
        shape : tuple of integers
            The surface shape.
        startswith1 : bool
            If True, columns and row starts with 1 instead of 0.
        topixel : Operator, optional
            World-to-pixel coordinate transform.
        to1d : Operator, optional
            Nd-to-1d pixel index transform.

        """
        origins = 'upper', 'lower'
        if not isinstance(origin, str):
            raise TypeError('Invalid origin.')
        origin = origin.lower()
        if origin not in origins:
            raise ValueError(
                'Invalid origin {0!r}. Expected values are {1}.'.format(
                    origin, strenum(origins)))
        Scene.__init__(self, shape, topixel=topixel, **keywords)
        if self.ndim != 2:
            raise ValueError('The scene is not 2-dimensional.')
        self.origin = origin
        self.startswith1 = bool(startswith1)
        if to1d is not None:
            to1d = asoperator(to1d)
            self.toNd = to1d.I
        else:
            self.toNd = None
        self.to1d = to1d

    @classmethod
    def fromfits(cls, header, **keywords):
        """
        Return a Scene grid described by the WCS of a FITS header.

        Parameters
        ----------
        header : astropy.io.fits.Header
            The FITS header that defines the WCS transform.

        """
        shape = fitsheader2shape(header)
        topixel = WCSToPixelOperator(header)
        to1d = To1dOperator(shape[::-1], order='f')
        scene = cls(shape, topixel, to1d, origin='lower', **keywords)
        scene.header = header
        return scene

    def column(self):
        return self.index % self.shape[1] + self.startswith1

    def row(self):
        return self.index // self.shape[1] + self.startswith1

    def get_integration_operator(self, polygons, ncolmax=0, dtype=np.float64,
                                 dtype_index=None, **keywords):
        """
        Return the Projection Operator that spatially integrates
        the plane pixel values enclosed by a sequence of polygons.

        Parameters
        ----------
        polygons : array-like of shape (..., nvertices, 2)
            The vertices defining the polygons inside which the spatial
            integration is done, in pixel coordinate units.
        ncolmax : integer
            The maximum number of surface pixel intersected by the polygons.
            If set to zero, or less than the required value, a two-pass
            computation will be performed to determine it.
        dtype : dtype, optional
            The float datatype of the matrix elements.
        dtype_index : dtype, optional
            The integer datatype of the matrix. The defaultis np.int32 or
            np.int64 depending on the number of elements of the scene.

        """
        if len(self) != len(self.all):
            raise NotImplementedError('The scene has removed elements.')
        dtype = np.dtype(dtype)
        if dtype_index is None:
            if len(self) <= np.iinfo(np.int32).max:
                dtype_index = np.int32
            else:
                dtype_index = np.int64
        dtype_index = np.dtype(dtype_index)
        if str(dtype_index) not in ('int32', 'int64') or \
           str(dtype) not in ('float32', 'float64'):
            raise TypeError(
                'The projection matrix cannot be created with types: {0} and {'
                '1}.'.format(dtype_index, dtype))

        polygons = np.array(polygons, dtype=dtype, copy=False)
        shape = polygons.shape[:-2]
        matrix = FSRMatrix((product(shape), product(self.shape)), dtype=dtype,
                           dtype_index=dtype_index, ncolmax=ncolmax)

        if matrix.data.size == 0:
            # f2py doesn't accept zero-sized opaque arguments
            data = np.empty(1, np.int8)
        else:
            data = matrix.data.ravel().view(np.int8)

        nvertices = polygons.shape[-2]
        func = 'matrix_polygon_integration_i{0}_r{1}'.format(
            dtype_index.itemsize, dtype.itemsize)
        min_ncolmax, outside = getattr(flib.projection, func)(
            polygons.reshape(-1, nvertices, 2).T, self.shape[1],
            self.shape[0], data, ncolmax)

        if min_ncolmax > ncolmax:
            return self.get_integration_operator(
                polygons, ncolmax=min_ncolmax, dtype=dtype,
                dtype_index=dtype_index, **keywords)

        out = ProjectionOperator(matrix, shapein=self.shape, shapeout=shape,
                                 **keywords)
        out.method = 'polygon integration'
        out.outside = bool(outside)
        out.min_ncolmax = min_ncolmax
        return out



from __future__ import absolute_import, division, print_function
from pyoperators.utils import strenum
from .operators import Spherical2HealpixOperator
from ...packedtables import Scene

__all__ = ['SceneHealpix']


class SceneHealpix(Scene):
    """
    Class for Healpix scenes.

    """
    def __init__(self, nside, kind='I', nest=False,
                 _convention='zenith,azimuth', **keywords):
        """
        nside : int
            Value of the map resolution parameter.*
        kind : 'I', 'QU' or 'IQU'
            The kind of sky: intensity-only, Q and U Stokes parameters only or
            intensity, Q and U.
        _convention : string, optional
            One of the following spherical coordinate conventions:
            'zenith,azimuth', 'azimuth,zenith', 'elevation,azimuth' and
            'azimuth,elevation'.
        nest : boolean, optional
            For the nested numbering scheme, set it to True. Default is
            the ring scheme.

        """
        kinds = 'I', 'QU', 'IQU'
        if not isinstance(kind, str):
            raise TypeError(
                'Invalid type {0!r} for the scene kind. Expected type is strin'
                'g.'.format(type(kind).__name__))
        kind = kind.upper()
        if kind not in kinds:
            raise ValueError('Invalid scene kind {0!r}. Expected kinds are: {1'
                             '}.'.format(kind, strenum(kinds)))
        nside = int(nside)
        topixel = Spherical2HealpixOperator(nside, _convention, nest)
        shape = (12 * nside**2,)
        if kind != 'I':
            shape += (len(kind),)
        Scene.__init__(self, shape, topixel, ndim=1, **keywords)
        self.kind = kind
        self.nside = nside
        self.nest = bool(nest)



from __future__ import division

import numpy as np
from numpy.testing import assert_almost_equal
from pyoperators.utils import product
from pyoperators.utils.testing import assert_same
from pysimulators.geometry import (
    convex_hull, create_circle, create_grid, create_grid_squares,
    create_rectangle, create_regular_polygon, create_square, rotate,
    surface_simple_polygon)

DTYPES = [np.float16, np.float32, np.float64, np.float128]
SHAPES = [(), (1,), (1, 1), (1, 2), (2, 1), (2, 2)]


def test_convex_hull():
    p = np.array([[1., 1], [1, -1], [-1, -1], [-1, 1]])
    expected = p[[0, 3, 2, 1]]
    points = [p, np.vstack((p, [[0, 0]])), np.vstack((p, [[-1, 1]]))]

    def func(point):
        h = convex_hull(point)
        assert_same(h, expected)
    for point in points:
        yield func, point


def test_rotate():
    e = [np.array(1, np.float128) / 2, np.sqrt(np.array(3, np.float128)) / 2]

    def func(dtype, shape):
        sqrt3 = np.sqrt(np.array(3, dtype))
        c = [sqrt3 / 2, np.array(1, dtype) / 2]
        coords = np.empty(shape + (2,), dtype)
        coords[...] = c
        assert_same(rotate(coords, 30), e, broadcasting=True)
        assert_same(rotate(coords, 30, out=coords), e, broadcasting=True)
    for dtype in DTYPES:
        for shape in SHAPES:
            yield func, dtype, shape


def test_circle():
    origin = [1, 1]

    def func(dtype, shape):
        n = max(product(shape), 1)
        radius = np.arange(1, n + 1, dtype=dtype).reshape(shape)
        circle = create_circle(radius, center=origin, dtype=dtype)
        actual = np.sqrt(np.sum((circle - origin)**2, axis=-1))
        assert_same(actual, radius[..., None], broadcasting=True)
    for dtype in DTYPES:
        for shape in SHAPES:
            yield func, dtype, shape


def test_grid():
    def create_grid_slow(shape, spacing, xreflection=False, yreflection=True,
                         center=(0, 0)):
        x = np.arange(shape[1], dtype=float) * spacing
        x -= x.mean()
        if xreflection:
            x = x[::-1]
        y = (np.arange(shape[0], dtype=float) * spacing)[::-1]
        y -= y.mean()
        if yreflection:
            y = y[::-1]
        grid_x, grid_y = np.meshgrid(x, y)
        return np.asarray([grid_x.T, grid_y.T]).T + center

    shapes = ((1, 1), (1, 2), (2, 1), (2, 2), (2, 3), (3, 2), (3, 3))
    spacing = 0.4
    origin = [1, 1]

    def func(shape, xreflection, yreflection):
        actual = create_grid(shape, spacing, xreflection=xreflection,
                             yreflection=yreflection, center=origin)
        expected = create_grid_slow(shape, spacing, xreflection=xreflection,
                                    yreflection=yreflection, center=origin)
        assert_same(actual, expected)
    for shape in shapes:
        for xreflection in (False, True):
            for yreflection in (False, True):
                yield func, shape, xreflection, yreflection


def test_grid_squares():

    def rotate_slow(x, angle, xreflection=False, yreflection=False):
        angle = np.radians(angle)
        r11 = np.cos(angle)
        r21 = np.sin(angle)
        r12 = -r21
        r22 = r11
        if xreflection:
            r11 = -r11
            r21 = -r21
        if yreflection:
            r12 = -r12
            r22 = -r22
        m = np.array([[r11, r12],
                      [r21, r22]])
        return np.dot(x, m.T)

    def create_grid_squares_slow(shape, spacing, filling_factor=1,
                                 xreflection=False, yreflection=False,
                                 center=(0, 0), angle=0):
        x = np.arange(shape[1], dtype=float) * spacing
        x -= x.mean()
        y = (np.arange(shape[0], dtype=float) * spacing)[::-1]
        y -= y.mean()
        grid_x, grid_y = np.meshgrid(x, y)
        nodes = np.asarray([grid_x.T, grid_y.T]).T
        d = np.sqrt(filling_factor * spacing**2) / 2
        offset = np.array([[d, d], [-d, d], [-d, -d], [d, -d]])
        corner = nodes[..., None, :] + offset
        corner = rotate_slow(corner, angle, xreflection, yreflection)
        return corner + center

    shapes = ((1, 1), (1, 2), (2, 1), (2, 2), (2, 3), (3, 2), (3, 3))
    spacing = 0.4
    filling_factor = 0.5
    origin = [1, 1]

    def func(shape, xreflection, yreflection):
        actual = create_grid_squares(
            shape, spacing, filling_factor=filling_factor,
            xreflection=xreflection, yreflection=yreflection, center=origin)
        expected = create_grid_squares_slow(
            shape, spacing, filling_factor=filling_factor,
            xreflection=xreflection, yreflection=yreflection, center=origin)
        assert_same(actual, expected)
    for shape in shapes:
        for xreflection in (False, True):
            for yreflection in (False, True):
                yield func, shape, xreflection, yreflection


def test_square():
    origin = np.asarray([1, 1])

    def func(dtype, shape):
        n = max(product(shape), 1)
        size = np.arange(1, n + 1, dtype=dtype).reshape(shape)
        actual = create_square(size, center=origin, dtype=dtype)
        expected = np.empty(shape + (4, 2))
        expected[..., 0, 0] = origin[0] + size / 2
        expected[..., 0, 1] = origin[1] + size / 2
        expected[..., 1, 0] = origin[0] - size / 2
        expected[..., 1, 1] = origin[1] + size / 2
        expected[..., 2, 0] = origin[0] - size / 2
        expected[..., 2, 1] = origin[1] - size / 2
        expected[..., 3, 0] = origin[0] + size / 2
        expected[..., 3, 1] = origin[1] - size / 2
        assert_same(actual, expected)
    for dtype in DTYPES:
        for shape in SHAPES:
            yield func, dtype, shape


def test_rectangle():
    origin = np.asarray([1, 1])

    def func(dtype, shape):
        n = max(product(shape), 1)
        size_x = np.arange(1, n + 1, dtype=dtype).reshape(shape)
        size_y = size_x / 2
        actual = create_rectangle(np.asarray([size_x.T, size_y.T]).T,
                                  center=origin, dtype=dtype)
        expected = np.empty(shape + (4, 2))
        expected[..., 0, 0] = origin[0] + size_x / 2
        expected[..., 0, 1] = origin[1] + size_y / 2
        expected[..., 1, 0] = origin[0] - size_x / 2
        expected[..., 1, 1] = origin[1] + size_y / 2
        expected[..., 2, 0] = origin[0] - size_x / 2
        expected[..., 2, 1] = origin[1] - size_y / 2
        expected[..., 3, 0] = origin[0] + size_x / 2
        expected[..., 3, 1] = origin[1] - size_y / 2
        assert_same(actual, expected)
    for dtype in DTYPES:
        for shape in SHAPES:
            yield func, dtype, shape


def test_surface_simple_polygon():
    origin = np.asarray([1, 1])
    nvertices = [3, 4, 5, 6, 7]
    expected = [n*np.sin(2*np.pi/n)/2 for n in nvertices]
    radius = np.array([1, 2])

    def func(r, n, e):
        polygon = create_regular_polygon(n, r, center=origin)
        assert_same(surface_simple_polygon(polygon), e * r**2)
        out = np.empty(np.shape(r))
        surface_simple_polygon(polygon, out=out)
        assert_same(out, e * r**2)
    for r in (radius[0], radius):
        for n, e in zip(nvertices, expected):
            yield func, r, n, e



"""
The version number is obtained from git tags, branch and commit identifier.
It has been designed for the following workflow:

- git init
- create setup.py commit
- more commit
- set version 0.1 in setup.py -> 0.1.dev03
- modify, commit              -> 0.1.dev04
- git checkout -b v0.1        -> 0.1.dev04
- modify, commit              -> 0.1.pre01
- modify, commit              -> 0.1.pre02
- git tag 0.1                 -> 0.1
- modify... and commit        -> 0.1.post01
- modify... and commit        -> 0.1.post02
- git tag 0.1.1               -> 0.1.1
- modify... and commit        -> 0.1.1.post01
- git checkout master         -> 0.1.dev04
- set version=0.2 in setup.py -> 0.2.dev01
- modify, commit              -> 0.2.dev02
- git tag 0.2                 -> 0.2
- set version=0.3 in setup.py -> 0.3.dev01


When working on the master branch, the dev number is the number of commits
since the last release branch (by default of name "v[0-9.]+", but it is
configurable) or the last tag.

"""

# These variables can be changed by the hooks importer
ABBREV = 5
F2PY_TABLE = {'integer': {'int8': 'char',
                          'int16': 'short',
                          'int32': 'int',
                          'int64': 'long_long'},
              'real': {'real32': 'float',
                       'real64': 'double'},
              'complex': {'real32': 'complex_float',
                          'real64': 'complex_double'}}
REGEX_RELEASE = '^v(?P<name>[0-9.]+)$'
try:
    import os
    from Cython.Build import cythonize
    USE_CYTHON = bool(int(os.getenv('SETUPHOOKS_USE_CYTHON', '1') or '0'))
except ImportError:
    USE_CYTHON = False

import numpy
import re
import shutil
import sys
from distutils.command.clean import clean
from numpy.distutils.command.build import build
from numpy.distutils.command.build_src import build_src
from numpy.distutils.command.sdist import sdist
from numpy.distutils.core import Command
from numpy.distutils.misc_util import has_f_sources
from subprocess import call, Popen, PIPE
from warnings import filterwarnings

try:
    root = os.path.dirname(os.path.abspath(__file__))
except NameError:
    root = os.path.dirname(os.path.abspath(sys.argv[0]))

# monkey patch to allow pure and elemental routines in preprocessed
# Fortran libraries
numpy.distutils.from_template.routine_start_re = re.compile(
    r'(\n|\A)((     (\$|\*))|)\s*((im)?pure\s+|elemental\s+)*(subroutine|funct'
    r'ion)\b', re.I)
numpy.distutils.from_template.function_start_re = re.compile(
    r'\n     (\$|\*)\s*((im)?pure\s+|elemental\s+)*function\b', re.I)


def get_cmdclass():

    class BuildCommand(build):
        def run(self):
            _write_version(self.distribution.get_name(),
                           self.distribution.get_version())
            build.run(self)

    class BuildSrcCommand(build_src):
        def initialize_options(self):
            build_src.initialize_options(self)
            self.f2py_opts = '--quiet'

        def run(self):
            has_fortran = False
            has_cython = False
            for ext in self.extensions:
                has_fortran = has_fortran or has_f_sources(ext.sources)
                for isource, source in enumerate(ext.sources):
                    if source.endswith('.pyx'):
                        if not USE_CYTHON:
                            ext.sources[isource] = source[:-3] + 'c'
                        else:
                            has_cython = True
            if has_fortran:
                with open(os.path.join(root, '.f2py_f2cmap'), 'w') as f:
                    f.write(repr(F2PY_TABLE))
            if has_cython:
                build_dir = None if self.inplace else self.build_src
                new_extensions = cythonize(self.extensions, force=True,
                                           build_dir=build_dir)
                for i in range(len(self.extensions)):
                    self.extensions[i] = new_extensions[i]
            build_src.run(self)

    class SDistCommand(sdist):
        def make_release_tree(self, base_dir, files):
            _write_version(self.distribution.get_name(),
                           self.distribution.get_version())
            initfile = os.path.join(self.distribution.get_name(),
                                    '__init__.py')
            new_files = []
            for f in files:
                if f.endswith('.pyx'):
                    new_files.append(f[:-3] + 'c')
            if initfile not in files:
                new_files.append(initfile)
            files.extend(new_files)
            sdist.make_release_tree(self, base_dir, files)

    class CleanCommand(clean):
        def run(self):
            clean.run(self)
            try:
                print(run_git('clean -fdX' + ('n' if self.dry_run else '')))
                return
            except RuntimeError:
                pass

            extensions = '.o', '.pyc', 'pyd', 'pyo', '.so'
            for root_, dirs, files in os.walk(root):
                for f in files:
                    if os.path.splitext(f)[-1] in extensions:
                        self.__delete(os.path.join(root_, f))
                for d in dirs:
                    if d in ('build', '__pycache__'):
                        self.__delete(os.path.join(root_, d), dir=True)
            files = (
                'MANIFEST',
                os.path.join(self.distribution.get_name(), '__init__.py'))
            for f in files:
                if os.path.exists(f):
                    self.__delete(f)

        def __delete(self, file_, dir=False):
            msg = 'would remove' if self.dry_run else 'removing'
            try:
                if not self.dry_run:
                    if dir:
                        shutil.rmtree(file_)
                    else:
                        os.unlink(file_)
            except OSError:
                msg = 'problem removing'
            print(msg + ' {!r}'.format(file_))

    class CoverageCommand(Command):
        description = "run the package coverage"
        user_options = [('file=', 'f', 'restrict coverage to a specific file'),
                        ('erase', None,
                         'erase previously collected coverage before run'),
                        ('html-dir=', None,
                         'Produce HTML coverage information in dir')]

        def run(self):
            cmd = [sys.executable, '-mnose', '--with-coverage', '--cover-html',
                   '--cover-package=' + self.distribution.get_name(),
                   '--cover-html-dir=' + self.html_dir]
            if self.erase:
                cmd.append('--cover-erase')
            call(cmd + [self.file])

        def initialize_options(self):
            self.file = 'test'
            self.erase = 0
            self.html_dir = 'htmlcov'

        def finalize_options(self):
            pass

    class TestCommand(Command):
        description = "run the test suite"
        user_options = [('file=', 'f', 'restrict test to a specific file')]

        def run(self):
            call([sys.executable, '-mnose', self.file])

        def initialize_options(self):
            self.file = 'test'

        def finalize_options(self):
            pass

    return {'build': BuildCommand,
            'build_src': BuildSrcCommand,
            'clean': CleanCommand,
            'coverage': CoverageCommand,
            'sdist': SDistCommand,
            'test': TestCommand}


def get_version(name, default):
    return _get_version_git(default) or _get_version_init_file(name) or default


def run_git(cmd, cwd=root):
    git = 'git'
    if sys.platform == 'win32':
        git = 'git.cmd'
    cmd = git + ' ' + cmd
    process = Popen(cmd.split(), cwd=cwd, stdout=PIPE, stderr=PIPE)
    stdout, stderr = process.communicate()
    if process.returncode != 0:
        if stderr != '':
            stderr = '\n' + stderr.decode('utf-8')
        raise RuntimeError(
            'Command failed (error {}): {}{}'.format(
                process.returncode, cmd, stderr))
    return stdout.strip().decode('utf-8')


def _get_version_git(default):
    INF = 2147483647

    def get_branches():
        return run_git('for-each-ref --sort=-committerdate --format=%(refname)'
                       ' refs/heads refs/remotes/origin').splitlines()

    def get_branch_name():
        branch = run_git('rev-parse --abbrev-ref HEAD')
        if branch != 'HEAD':
            return branch
        branch = run_git('branch --no-color --contains HEAD').splitlines()
        return branch[min(1, len(branch)-1)].strip()

    def get_description():
        branch = get_branch_name()
        try:
            description = run_git('describe --abbrev={} --tags'.format(ABBREV))
        except RuntimeError:
            description = run_git(
                'describe --abbrev={} --always'.format(ABBREV))
            regex = r"""^
            (?P<commit>.*?)
            (?P<dirty>(-dirty)?)
            $"""
            m = re.match(regex, description, re.VERBOSE)
            commit, dirty = (m.group(_) for _ in 'commit,dirty'.split(','))
            return branch, '', INF, commit, dirty

        regex = r"""^
        (?P<tag>.*?)
        (?:-
            (?P<rev>\d+)-g
            (?P<commit>[0-9a-f]{5,40})
        )?
        (?P<dirty>(-dirty)?)
        $"""
        m = re.match(regex, description, re.VERBOSE)
        tag, rev, commit, dirty = (m.group(_)
                                   for _ in 'tag,rev,commit,dirty'.split(','))
        if rev is None:
            rev = 0
            commit = ''
        else:
            rev = int(rev)
        return branch, tag, rev, commit, dirty

    def get_rev_since_branch(branch):
        try:
            # get best common ancestor
            common = run_git('merge-base HEAD ' + branch)
        except RuntimeError:
            return INF  # no common ancestor, the branch is dangling
        return int(run_git('rev-list --count HEAD ^' + common))

    def get_rev_since_any_branch():
        if REGEX_RELEASE.startswith('^'):
            regex = REGEX_RELEASE[1:]
        else:
            regex = '.*' + REGEX_RELEASE
        regex = '^refs/(heads|remotes/origin)/' + regex

        branches = get_branches()
        for branch in branches:
            # filter branches according to BRANCH_REGEX
            if not re.match(regex, branch):
                continue
            rev = get_rev_since_branch(branch)
            if rev > 0:
                return rev
        # no branch has been created from an ancestor
        return INF

    try:
        run_git('rev-parse --is-inside-work-tree')
    except (OSError, RuntimeError):
        return ''

    branch, tag, rev_tag, commit, dirty = get_description()

    # check if the commit is tagged
    if rev_tag == 0:
        return tag + dirty

    # if the current branch is master, look up the closest tag or the closest
    # release branch rev to get the dev number otherwise, look up the closest
    # tag or the closest master rev.
    suffix = 'dev'
    if branch == 'master':
        rev_branch = get_rev_since_any_branch()
        name = default
        is_branch_release = False
    else:
        rev_branch = get_rev_since_branch('master')
        name = branch
        m = re.match(REGEX_RELEASE, branch)
        is_branch_release = m is not None
        if is_branch_release:
            try:
                name = m.group('name')
            except IndexError:
                pass
        elif rev_tag == rev_branch:
            tag = branch

    if rev_branch == rev_tag == INF:
        # no branch and no tag from ancestors, counting from root
        rev = int(run_git('rev-list --count HEAD'))
        if branch != 'master':
            suffix = 'rev'
    elif rev_tag <= rev_branch:
        rev = rev_tag
        if branch != 'master':
            name = tag
        if is_branch_release:
            suffix = 'post'
    else:
        rev = rev_branch
        if is_branch_release:
            suffix = 'pre'
    if name != '':
        name += '.'
    return '{}{}{:02}{}'.format(name, suffix, rev, dirty)


def _get_version_init_file(name):
    try:
        f = open(os.path.join(name, '__init__.py')).read()
    except IOError:
        return ''
    m = re.search(r"__version__ = '(.*)'", f)
    if m is None:
        return ''
    return m.groups()[0]


def _write_version(name, version):
    try:
        init = open(os.path.join(root, name, '__init__.py.in')).readlines()
    except IOError:
        return
    init += ['\n', '__version__ = ' + repr(version) + '\n']
    open(os.path.join(root, name, '__init__.py'), 'w').writelines(init)


filterwarnings('ignore', "Unknown distribution option: 'install_requires'")



import numpy as np
import os

from astropy.io import fits as pyfits
from astropy.wcs import WCS
from pyoperators.utils.testing import assert_eq, assert_same, skiptest_unless_module
from pysimulators.wcsutils import (
    angle_lonlat, barycenter_lonlat, combine_fitsheader, create_fitsheader,
    create_fitsheader_for, fitsheader2shape, get_cdelt_pa, has_wcs, mean_degrees,
    WCSToWorldOperator)


def test_mean_degrees():
    assert mean_degrees([1, 2]) == 1.5
    assert_same(mean_degrees([1, 359.1]), 0.05, atol=100)
    assert_same(mean_degrees([0.1, 359.1]), 359.6)


def test_angle_lonlat1():
    assert_same(angle_lonlat(30, 0, 40, 0), 10, atol=100)


def test_angle_lonlat2():
    input = (((30, 0), (40, 0), 10),
             ((39, 0), (92, 90), 90),
             ((39, 0), (37, -90), 90),
             ((37, -90), (39, 0), 90),
             ((100, 90), (134, -90), 180),
             ((24, 30), (24, 32), 2))

    def func(c1, c2, angle):
        assert_same(angle_lonlat(c1, c2), angle, rtol=1000)
    for c1, c2, angle in input:
        yield func, c1, c2, angle


def test_barycenter_lonlat():
    assert_eq(barycenter_lonlat([30, 40], [0, 0]), (35, 0))
    assert_eq(barycenter_lonlat([20, 20, 20], [-90, 0, 90]), (20, 0))
    assert_eq(barycenter_lonlat([20, 20, 20], [0, 45, 90]), (20, 45))


def test_get_cdelt_pa1():
    header = create_fitsheader((10, 10), cdelt=(-1.2, 3))
    cdelt, pa = get_cdelt_pa(header)
    assert_eq(cdelt, (-1.2, 3))
    assert_eq(pa, 0)


def test_get_cdelt_pa2():
    cdelt = (-1.5, 3)
    pa = -25.
    header = create_fitsheader((10, 10), cdelt=cdelt, pa=pa)
    cdelt_, pa_ = get_cdelt_pa(header)
    assert_eq(cdelt, cdelt_)
    assert_eq(pa, pa_)


def test_combine_fitsheader():
    headers = [
        create_fitsheader((1, 1), cdelt=3., crval=(0, 0)),
        create_fitsheader((3, 3), cdelt=1., crval=(1, 1)),
        create_fitsheader((5, 5), cdelt=1., crval=(3, 3)),
        create_fitsheader((2, 2), cdelt=1., crval=(5, 5)),
    ]
    header0 = combine_fitsheader(headers)
    proj0 = WCS(header0)
    epsilon = 1.e-10

    def func(iheader, header):
        nx = header['NAXIS1']
        ny = header['NAXIS2']
        x = header['CRPIX1']
        y = header['CRPIX2']
        edges = (np.array(3*(0.5, x, nx+0.5)),
                 np.array((0.5, 0.5, 0.5, y, y, y, ny+0.5, ny+0.5, ny+0.5)))
        a, d = WCS(header).wcs_pix2world(edges[0], edges[1], 1)
        x0, y0 = proj0.wcs_world2pix(a, d, 1)
        assert np.all(x0 > 0.5-epsilon)
        assert np.all(y0 > 0.5-epsilon)
        assert np.all(x0 < header0['NAXIS1'] + 0.5 + epsilon)
        assert np.all(y0 < header0['NAXIS2'] + 0.5 + epsilon)
        if iheader == 0:
            assert y0[0] <= 1.5 and x0[-1] >= header0['NAXIS1']-0.5
        if iheader == len(headers)-1:
            assert x0[0] <= 1.5 and y0[-1] >= header0['NAXIS2']-0.5
    for iheader, header in enumerate(headers):
        yield func, iheader, header


def test_has_wcs():
    header = create_fitsheader([2, 4])
    assert not has_wcs(header)
    header = create_fitsheader([2, 4], cdelt=1)
    assert has_wcs(header)


def test_create_fitsheader1():
    shapes = [(), (1,), (1, 2), (0, 1), (1, 0), (2, 3)]
    arrays = [np.ones(s) for s in shapes]

    def func(a):
        h1 = create_fitsheader(a.shape[::-1],
                               dtype=float if a.ndim > 0 else np.int32)
        h2 = create_fitsheader_for(a if a.ndim > 0 else None)
        assert h1 == h2
        assert h1['NAXIS'] == a.ndim
        for i, s in enumerate(a.shape[::-1]):
            assert h1['NAXIS'+str(i+1)] == s
    for a in arrays:
        yield func, a


def test_create_fitsheader2():
    header = create_fitsheader_for(None)
    assert header['NAXIS'] == 0


def test_create_fitsheader3():
    header = create_fitsheader((10, 3), cdelt=1)
    assert_eq(header['RADESYS'], 'ICRS')
    assert 'EQUINOX' not in header


def test_fitsheader2shape():
    def func(naxes):
        header = create_fitsheader(naxes)
        assert_eq(fitsheader2shape(header), naxes[::-1])
    for naxes in [(), (1,), (1, 2), (2, 0, 3)]:
        yield func, naxes


@skiptest_unless_module('kapteyn')
def test_wcsoperator_kapteyn():
    from pysimulators.wcsutils import WCSKapteynToWorldOperator
    path = os.path.join(os.path.dirname(__file__), 'data/header_gnomonic.fits')
    header = pyfits.open(path)[0].header

    toworld_kapteyn = WCSKapteynToWorldOperator(header)
    crpix = (header['CRPIX1'], header['CRPIX2'])
    crval = (header['CRVAL1'], header['CRVAL2'])
    assert_eq(toworld_kapteyn(crpix), crval)
    assert_eq(toworld_kapteyn.I(crval), crpix)


def test_wcsoperator():
    path = os.path.join(os.path.dirname(__file__), 'data/header_gnomonic.fits')
    header = pyfits.open(path)[0].header

    def func(origin):
        crval = (header['CRVAL1'], header['CRVAL2'])
        crpix = np.array((header['CRPIX1'], header['CRPIX2'])) - 1 + origin
        toworld = WCSToWorldOperator(header, origin)
        assert_same(toworld(crpix), crval, atol=10)
        assert_same(toworld.I(crval), crpix, atol=100000)

    for origin in (0, 1):
        yield func, origin



from __future__ import division

import numpy as np
from numpy.testing import assert_equal

from pysimulators import PointingMatrix


def test_constructor():
    shape = (2,)
    shape_input = (3, 4)
    array = np.zeros(shape, dtype=np.int64)

    def func(matrix):
        assert matrix.shape_input == shape_input
        assert matrix.value.shape == shape
        assert matrix.index.shape == shape

    for matrix in [PointingMatrix(array, shape_input, copy=False),
                   PointingMatrix.empty(shape, shape_input),
                   PointingMatrix.zeros(shape, shape_input)]:
        yield func, matrix


# we assume that the input map is 2x5
#  -----------
#  |0|1|2|3|4|
#  |-+-+-+-+-|
#  |5|6|7|8|9|
#  -----------
# The first sample intersects a quarter of 0, 1, 5 and 6
# and the second sample intersects half of 6 and 7.
matrix = PointingMatrix.empty((2, 4), (2, 5))
matrix[0] = [(0, .25), (1, .25), (5, .25), (6, .25)]
matrix[1] = [(7, .50), (8, .50), (-1, 0),  (-1, 0)]


def test_isvalid():
    assert matrix.isvalid()

    invalid1 = matrix.copy()
    invalid1[0, 2] = (-2, 0)
    assert not invalid1.isvalid()

    invalid2 = matrix.copy()
    invalid2[0, 2] = (10, 0)
    assert not invalid2.isvalid()


def test_mask():
    expected = [[False, False, True, True, True],
                [False, False, False, False, True]]
    assert_equal(matrix.get_mask(), expected)


def test_pack():
    expected = PointingMatrix.empty((2, 4), (6,))
    expected[0] = [(0, .25), (1, .25), (2, .25), (3, .25)]
    expected[1] = [(4, .50), (5, .50), (-1, 0),  (-1, 0)]
    packed = matrix.copy()
    packed.pack(matrix.get_mask())
    assert_equal(packed, expected)
    assert_equal(packed.shape_input, expected.shape_input)



#!/usr/bin/env python
import numpy as np
import hooks
import sys

from distutils.util import get_platform
from numpy.distutils.core import setup
from numpy.distutils.extension import Extension
from hooks import get_cmdclass, get_version

hooks.F2PY_TABLE = {
    'integer': {'int8': 'char',
                'int16': 'short',
                'int32': 'int',
                'int64': 'long_long'},
    'real': {'sp': 'float',
             'dp': 'double',
             'p': 'double',
             'real32': 'float',
             'real64': 'double'},
    'complex': {'sp': 'complex_float',
                'dp': 'complex_double',
                'p': 'complex_double',
                'real32': 'complex_float',
                'real64': 'complex_double'}}

VERSION = '1.1'

name = 'pysimulators'
long_description = open('README.rst').read()
keywords = 'scientific computing'
platforms = 'MacOS X,Linux,Solaris,Unix,Windows'
define_macros = [('GFORTRAN', None), ('PRECISION_REAL', 8)]
if sys.platform == 'darwin':
    extra_f90_compile_args = ['-O2 -funroll-loops -cpp',
                              '-fopenmp -fpack-derived -Wall -g']
else:
    extra_f90_compile_args = ['-Ofast -funroll-loops -march=native -cpp',
                              '-fopenmp -fpack-derived -Wall -g']

# debugging options: -fcheck=all -fopt-info-vec-missed
# -fprofile-use -fprofile-correction
# ifort: -xHost -vec-report=1
mod_dir = 'build/temp.' + get_platform() + '-%s.%s' % sys.version_info[:2]

flib = ('fmod',
        {'sources': ['src/module_precision.f90',
                     'src/module_tamasis.f90',
                     'src/module_string.f90',
                     'src/module_fitstools.f90',
                     'src/module_geometry.f90.src',
                     'src/module_math.f90.src',
                     'src/module_math_old.f90',
                     'src/module_pointingmatrix.f90',
                     'src/module_operators.f90.src',
                     'src/module_sort.f90',
                     'src/module_wcs.f90'],
         'depends': [],
         'macros': define_macros,
         'extra_f90_compile_args': extra_f90_compile_args,
         'include_dirs': [np.get_include()]})

ext_modules = [Extension('pysimulators._flib',
                         sources=['src/datautils.f90',
                                  'src/geometry.f90',
                                  'src/operators.f90.src',
                                  'src/pointingmatrix_old.f90',
                                  'src/projection.f90.src',
                                  'src/sparse.f90.src',
                                  'src/wcsutils.f90'],
                         define_macros=define_macros,
                         extra_compile_args=['-Wno-format'],
                         extra_f90_compile_args=extra_f90_compile_args,
                         include_dirs=[np.get_include(), mod_dir],
                         libraries=['gomp', flib])]


setup(name=name,
      version=get_version(name, VERSION),
      description='Tools to build an instrument model.',
      long_description=long_description,
      url='http://pchanial.github.com/pysimulators',
      author='Pierre Chanial',
      author_email='pierre.chanial@gmail.com',
      maintainer='Pierre Chanial',
      maintainer_email='pierre.chanial@gmail.com',
      install_requires=['pyoperators>=0.12.15', 'astropy>=0.3.2'],
      packages=['pysimulators',
                'pysimulators/interfaces',
                'pysimulators/interfaces/healpy',
                'pysimulators/interfaces/madmap1',
                'pysimulators/packedtables'],
      platforms=platforms.split(','),
      keywords=keywords.split(','),
      cmdclass=get_cmdclass(),
      ext_modules=ext_modules,
      license='CeCILL-B',
      classifiers=[
          'Programming Language :: Python',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.6',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.3',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Fortran',
          'Development Status :: 5 - Production/Stable',
          'Intended Audience :: Science/Research',
          'Operating System :: OS Independent',
          'Topic :: Scientific/Engineering',
          'Topic :: Scientific/Engineering :: Astronomy',
          'Topic :: Scientific/Engineering :: Atmospheric Science',
          'Topic :: Scientific/Engineering :: Information Analysis',
          'Topic :: Scientific/Engineering :: Medical Science Apps.',
          'Topic :: Scientific/Engineering :: Physics'])



from __future__ import division

import numpy as np

from nose.tools import assert_equal, assert_raises
from numpy.testing import assert_almost_equal, assert_array_equal
from pyoperators.utils.testing import assert_eq, assert_is, assert_is_instance
from pysimulators.quantities import Quantity, UnitError


def assert_quantity(q, m, u):
    assert_is_instance(q, Quantity)
    assert_array_equal(q.magnitude, m)
    if isinstance(u, dict):
        assert_equal(q._unit, u)
    else:
        assert_equal(q.unit, u)


def test1():
    q = Quantity(1, 'km')
    assert_eq(q.SI, Quantity(1000, 'm'))

    q = Quantity(1, 'm')
    assert_eq(q, q.tounit('m'))
    q2 = q.copy()
    q.inunit('m')
    assert_eq(q, q2)


def test2():
    q = Quantity(1, 'km').tounit('m')
    assert_quantity(q, 1000, 'm')


def test3():
    q = Quantity(1, 'km')
    q.inunit('m')
    assert_quantity(q, 1000, 'm')


def check_unit_add(x, y, v, u):
    c = x + y
    assert_quantity(c, v, u)


def test_add1():
    q = Quantity(1.)
    for other in Quantity(1.), np.array(1.), 1:
        yield check_unit_add, q, other, 2, {}
        yield check_unit_add, other, q, 2, {}


def test_add2():
    q = Quantity(1., 'm')
    for other in Quantity(1.), np.array(1.), 1:
        yield check_unit_add, q, other, 2, {'m': 1.0}
        yield check_unit_add, other, q, 2, {'m': 1.0}


def test_add3():
    q = Quantity(1.)
    q2 = Quantity(1., 'km')
    yield check_unit_add, q, q2, 2, {'km': 1.0}


def test_add4():
    q = Quantity(1., 'm')
    q2 = Quantity(1., 'km')
    yield check_unit_add, q, q2, 1001, {'m': 1.0}


def test_add5():
    q = Quantity(1., 'km')
    q2 = Quantity(1., 'm')
    yield check_unit_add, q, q2, 1.001, {'km': 1.0}


def test_add6():
    assert_raises(UnitError, lambda: Quantity(1, 'kloug') +
                  Quantity(3., 'babar'))


def check_unit_sub(x, y, v, u):
    c = x - y
    assert_quantity(c, v, u)


def test_sub1():
    q = Quantity(1.)
    for other in Quantity(1.), np.array(1.), 1:
        yield check_unit_sub, q, other, 0, {}
        yield check_unit_sub, other, q, 0, {}


def test_sub2():
    q = Quantity(1., 'm')
    for other in Quantity(1.), np.array(1.), 1:
        yield check_unit_sub, q, other, 0, {'m': 1.0}
        yield check_unit_sub, other, q, 0, {'m': 1.0}


def test_sub3():
    q = Quantity(1.)
    q2 = Quantity(1., 'km')
    yield check_unit_sub, q, q2, 0, {'km': 1.0}


def test_sub4():
    q = Quantity(1., 'm')
    q2 = Quantity(1., 'km')
    yield check_unit_sub, q, q2, -999, {'m': 1.0}


def test_sub5():
    q = Quantity(1., 'km')
    q2 = Quantity(1., 'm')
    yield check_unit_sub, q, q2, 0.999, {'km': 1.0}


def test_sub6():
    assert_raises(UnitError, lambda: Quantity(1, 'kloug') -
                  Quantity(3., 'babar'))


def test_conversion_error():
    a = Quantity(1., 'kloug')
    assert_raises(UnitError, lambda: a.inunit('notakloug'))
    assert_raises(UnitError, lambda: a.inunit('kloug^2'))


def test_conversion_sr1():
    a = Quantity(1, 'MJy/sr').tounit('uJy/arcsec^2')
    assert_almost_equal(a, 23.5044305391)
    assert_equal(a.unit, 'uJy / arcsec^2')


def test_conversion_sr2():
    a = (Quantity(1, 'MJy/sr')/Quantity(1, 'uJy/arcsec^2')).SI
    assert_almost_equal(a, 23.5044305391)
    assert_equal(a.unit, '')


def test_array_prepare():
    assert Quantity(10, 'm') <= Quantity(1, 'km')
    assert Quantity(10, 'm') < Quantity(1, 'km')
    assert Quantity(1, 'km') >= Quantity(10, 'm')
    assert Quantity(1, 'km') > Quantity(10, 'm')
    assert Quantity(1, 'km') != Quantity(1, 'm')
    assert Quantity(1, 'km') == Quantity(1000, 'm')
    assert np.maximum(Quantity(10, 'm'), Quantity(1, 'km')) == 1000
    assert np.minimum(Quantity(10, 'm'), Quantity(1, 'km')) == 10


def test_function():
    arrays = ([1.3, 2, 3], [[1., 2, 3], [4, 5, 6]])

    def func(array, f):
        a = Quantity(array, unit='Jy')
        b = f(a)
        if f not in (np.round,):
            assert np.isscalar(b)
        assert_array_equal(b, f(a.view(np.ndarray)))
        if f in (np.round,) or a.ndim == 1:
            return
        b = f(a, axis=0)
        assert_array_equal(b, f(a.view(np.ndarray), axis=0))
        assert_equal(b.unit, 'Jy' if f is not np.var else 'Jy^2')
    for array in arrays:
        for f in (np.min, np.max, np.mean, np.ptp, np.round, np.sum, np.std,
                  np.var):
            yield func, array, f


def test_dtype():
    assert_is(Quantity(1).dtype, np.dtype(float))
    assert_is(Quantity(1, dtype='float32').dtype, np.dtype(np.float32))
    assert_is(Quantity(1.).dtype, np.dtype(float))
    assert_is(Quantity(complex(1, 0)).dtype, np.dtype(np.complex128))
    assert_is(Quantity(1., dtype=np.complex64).dtype, np.dtype(np.complex64))
    assert_is(Quantity(1., dtype=np.complex128).dtype, np.dtype(np.complex128))
    assert_is(Quantity(1., dtype=np.complex256).dtype, np.dtype(np.complex256))
    assert_is(Quantity(np.array(complex(1, 0))).dtype, np.dtype(complex))
    assert_is(Quantity(np.array(np.complex64(1.))).dtype,
              np.dtype(np.complex64))
    assert_is(Quantity(np.array(np.complex128(1.))).dtype, np.dtype(complex))


def test_derived_units1():
    du = {'detector[rightward]': Quantity([1, 1/10.], 'm^2'),
          'time[leftward]': Quantity([1, 1/2, 1/3, 1/4], 's')}
    a = Quantity([[1, 2, 3, 4],
                  [10, 20, 30, 40]], 'detector', derived_units=du)
    assert_almost_equal(a.SI, [[1, 2, 3, 4], [1, 2, 3, 4]])
    a = Quantity([[1, 2, 3, 4], [10, 20, 30, 40]], 'time', derived_units=du)
    assert_almost_equal(a.SI, [[1, 1, 1, 1], [10, 10, 10, 10]])
    a = Quantity(1, 'detector C', du)
    assert a.SI.shape == (2,)
    a = Quantity(np.ones((1, 10)), 'detector', derived_units=du)
    assert a.SI.shape == (2, 10)


def test_derived_units2():
    a = Quantity(4., 'Jy/detector', {'detector': Quantity(2,'arcsec^2')})
    a.inunit(a.unit + ' / arcsec^2 * detector')
    assert_quantity(a, 2, 'Jy / arcsec^2')


def test_derived_units3():
    a = Quantity(2., 'brou', {'brou': Quantity(2., 'bra'),
                              'bra': Quantity(2., 'bri'),
                              'bri': Quantity(2., 'bro'),
                              'bro': Quantity(2, 'bru'),
                              'bru': Quantity(2., 'stop')})
    b = a.tounit('bra')
    assert b.magnitude == 4
    b = a.tounit('bri')
    assert b.magnitude == 8
    b = a.tounit('bro')
    assert b.magnitude == 16
    b = a.tounit('bru')
    assert b.magnitude == 32
    b = a.tounit('stop')
    assert b.magnitude == 64
    b = a.SI
    assert b.magnitude == 64
    assert b.unit == 'stop'


def test_derived_units4():
    l = np.arange(3*4*5).reshape((3, 4, 5))
    a_leftward = Quantity.ones((2, 3, 4, 5), 'r', {'r[leftward]': Quantity(l)})
    r = np.arange(2*3*4).reshape((2, 3, 4))
    a_rightward = Quantity.ones((2, 3, 4, 5), 'r',
                                {'r[rightward]': Quantity(r)})
    _ = slice(None)

    def func_leftward(a, key):
        b = a[key]
        if np.isscalar(b):
            assert b.dtype == np.float64
            return
        if key is Ellipsis:
            assert_eq(b.derived_units['r[leftward]'],
                      a.derived_units['r[leftward]'])
        if key in ((1, 2, 3, 4), (_, 1, 2, 3)):
            assert 'r' in b.derived_units
            if key == (1, 2, 3, 4):
                assert b.derived_units['r'] == Quantity(59)
            else:
                assert b.derived_units['r'] == Quantity(33)
        else:
            assert 'r' not in b.derived_units
            if not isinstance(key, tuple):
                key = (key,)
            key += (4-len(key)) * (slice(None),)
            assert_eq(b.derived_units['r[leftward]'],
                      a.derived_units['r[leftward]'][key[-3:]])

    def func_rightward(a, key):
        b = a[key]
        if np.isscalar(b):
            assert b.dtype == np.float64
            return
        if key is Ellipsis:
            assert_eq(b.derived_units['r[rightward]'],
                      a.derived_units['r[rightward]'])
        if key in ((1, 2, 3), (1, 2, 3, 4), (1, 2, 3, _)):
            assert 'r' in b.derived_units
            assert b.derived_units['r'] == Quantity(23)
        else:
            assert 'r' not in b.derived_units
            if not isinstance(key, tuple):
                key = (key,)
            assert_eq(b.derived_units['r[rightward]'],
                      a.derived_units['r[rightward]'][key[:3]])
    for key in (1, (1,), (1, 2), (1, 2, 3), (1, 2, 3, 4),
                _, (_,), (_, 1), (_, 1, 2), (_, 1, 2, 3),
                (1, _), (1, _, 2), (1, _, 2, 3),
                (1, 2, _), (1, 2, _, 3),
                (1, 2, 3, _),
                (_, _),  (_, _, 1), (_, 1, _), (1, _, _),
                (_, _, 1, 2), (_, 1, _, 2), (_, 1, 2, _), (1, _, _, 2),
                (1, _, 2, _), (1, 2, _, _),
                (_, _, _), (_, _, _, 1), (_, _, 1, _), (_, 1, _, _),
                (1, _, _, _),
                (_, _, _, _), Ellipsis):
        yield func_leftward, a_leftward, key
        yield func_rightward, a_rightward, key


def test_pixels():
    assert_quantity(Quantity(1, 'pixel/sr/pixel_reference').SI, 1, 'sr^-1')



"""
Patch PyOperator's SparseOperator to include the following formats:
    - Fixed Sparse Column (FSCMatrix)
    - Fixed Sparse Row (FSRMatrix)
    - Fixed Sparse Column Rotation 2d (FSCRotation2dMatrix)
    - Fixed Sparse Row Rotation 2d (FSRRotation2dMatrix)
    - Fixed Sparse Column Rotation 3d (FSCRotation3dMatrix)
    - Fixed Sparse Row Rotation 3d (FSRRotation3dMatrix)

"""
from __future__ import absolute_import, division, print_function
from pyoperators import (
    CompositionOperator, DiagonalOperator, operation_assignment)
from pyoperators.linear import SparseBase
from pyoperators.memory import empty
from pyoperators.utils import isscalarlike, product, tointtuple
from pysimulators._flib import sparse as fsp
from pysimulators._flib import operators as fop
import numpy as np
import operator
import pyoperators
import scipy.sparse as sp
import sys
if sys.version_info.major == 2:
    range = xrange

__all__ = []


class _FSMatrix(object):
    def __init__(self, flib_id, shape, block_shape, nmax, sparse_axis,
                 dtype=None, dtype_index=None, data=None, verbose=False):
        self._flib_id = flib_id
        if not isinstance(shape, (list, tuple)):
            raise TypeError("Invalid shape '{0}'.".format(shape))
        if len(shape) != 2:
            raise ValueError("The number of dimensions is not 2.")
        if len(block_shape) != 2:
            raise ValueError(
                "The number of dimensions of the blocks is not 2.")
        straxes = ('row', 'column')
        if sparse_axis == 1:
            straxes = straxes[::-1]
        if any(s % b != 0 for s, b in zip(shape, block_shape)):
            raise ValueError(
                "The shape of the matrix '{0}' is incompatible with blocks of "
                "shape '{1}'.".format(shape, block_shape))
        if data is None:
            if nmax is None:
                raise ValueError('The maximum number of non-zero {0}s per {1} '
                                 'is not specified.'.format(*straxes))
            shape_data = (shape[1-sparse_axis] // block_shape[0], nmax)
            if dtype is None:
                dtype = float
            if dtype_index is None:
                dtype_index = int
            dtype_data = self._get_dtype_data(dtype, dtype_index, block_shape)
            data = empty(shape_data, dtype_data, verbose=verbose)
        elif 'index' not in data.dtype.names:
            raise TypeError('The structured array has no field index.')
        elif len(data.dtype.names) == 1:
            raise TypeError('The structured array has no data field.')
        elif product(data.shape[:-1]) * block_shape[1-sparse_axis] != \
             shape[1-sparse_axis]:
            raise ValueError(
                "The shape of the matrix '{0}' is incompatible with that of th"
                "e structured array '{1}'.".format(shape, data.shape))
        elif nmax not in (None, data.shape[-1]):
            raise ValueError(
                "The n{0}max keyword value '{1}' is incompatible with the shap"
                "e of the structured array '{2}'.".format(
                    straxes[0][:3], nmax, data.shape))
        else:
            dtype_index = data.dtype['index']
            dtype = data.dtype[1]
            if dtype.type == np.void:
                dtype = dtype.subdtype[0].type
            expected = self._get_dtype_data(dtype, dtype_index, block_shape)
            if data.dtype != expected:
                raise TypeError(
                    'The input dtype {0} is invalid. Expected dtype is {1}.'.
                    format(data.dtype, expected))
            if nmax is None:
                nmax = data.shape[-1]
        self.dtype = np.dtype(dtype)
        self.data = data.view(np.recarray)
        self.ndim = 2
        self.shape = tuple(shape)
        self.block_shape = tuple(block_shape)
        setattr(self, 'n' + straxes[0][:3] + 'max', nmax)

    def _matvec(self, v, out, nmax):
        v = np.asarray(v).ravel()
        self._validatein(v.shape)
        if out is None:
            out = np.zeros(v.size * self.shape[0] // self.shape[1],
                           np.find_common_type([self.dtype, v.dtype], []))
        elif not isinstance(out, np.ndarray):
            raise TypeError('The output array is not an ndarray.')
        elif not out.flags.contiguous:
            raise ValueError('The output array is not contiguous.')
        elif out.size != v.size * self.shape[0] // self.shape[1]:
            raise ValueError(
                "The output size '{0}' is incompatible with the number of rows"
                " of the sparse matrix '{1}'.".format(out.size, self.shape[0]))
        else:
            out = out.ravel()
            self._validateout(out.shape)

        di = self.data.index.dtype
        ds = self.dtype
        dv = v.dtype
        if str(ds) not in ('float32', 'float64') or \
           str(di) not in ('int32', 'int64'):
            return v, out, False

        if dv.kind != 'f' or dv.type in (np.float16, np.float128) or \
           dv.type is np.float32 and ds.type is np.float64:
            v = v.astype(self.dtype)
            dv = self.dtype
        if out.dtype != dv or not out.flags.contiguous:
            out_ = np.empty(out.size, dtype=dv)
            out_[...] = out
        else:
            out_ = out
        flib_id = self._flib_id
        if isinstance(self, (FSCMatrix, FSRMatrix)):
            extra_size = v.size // self.shape[1]
            if extra_size > 1:
                flib_id += '_homothety'
        f = '{0}_matvec_i{1}_r{2}_v{3}'.format(
            flib_id, di.itemsize, ds.itemsize, dv.itemsize)
        func = getattr(fsp, f)
        m = self.data.ravel().view(np.int8)
        if flib_id.endswith('_homothety'):
            func(m, v, out_, nmax, self.shape[1], self.shape[0], extra_size)
        elif flib_id.endswith('_block'):
            func(m, v, out_, nmax, self.shape[1] // self.block_shape[1],
                 self.shape[0] // self.block_shape[0], self.block_shape[0],
                 self.block_shape[1])
        else:
            func(m, v, out_, nmax)
        if out.dtype != dv:
            out[...] = out_
        return v, out, True

    def __mul__(self, other):
        if isscalarlike(other):
            data = self.data.copy()
            data.value *= other
            return type(self)(self.shape, data=data)
        other = np.asarray(other)
        if other.ndim == 1:
            return self._matvec(other)
        return NotImplemented

    def __rmul__(self, other):
        if isscalarlike(other):
            return self * other
        other = np.asarray(other)
        if other.ndim == 1:
            return self._transpose() * other
        return NotImplemented

    def copy(self):
        return type(self)(self.shape, data=self.data.copy())

    def _reshapein(self, shape):
        extra_size = product(shape) // self.shape[1]
        if extra_size > 1 or shape[-1] == 1:
            return self.shape[0], extra_size
        return self.shape[0]

    def _reshapeout(self, shape):
        extra_size = product(shape) // self.shape[0]
        if extra_size > 1 or shape[-1] == 1:
            return self.shape[1], extra_size
        return self.shape[1]

    def _validatein(self, shape):
        size = product(shape)
        if not isinstance(self, (FSCMatrix, FSRMatrix)):
            if size == self.shape[1]:
                return
        elif size % self.shape[1] == 0:
            return
        raise ValueError(
            "The shape '{0}' is incompatible with the number of columns of the"
            " sparse matrix '{1}'.".format(shape, self.shape[1]))

    def _validateout(self, shape):
        size = product(shape)
        if not isinstance(self, (FSCMatrix, FSRMatrix)):
            if size == self.shape[0]:
                return
        elif size % self.shape[0] == 0:
            return
        raise ValueError(
            "The shape '{0}' is incompatible with the number of rows of the sp"
            "arse matrix '{1}'.".format(shape, self.shape[0]))

    def _get_dtype_data(self, dtype, dtype_index, block_shape):
        return [('index', dtype_index), ('value', dtype)]

    __array_priority__ = 10.1


class FSCMatrix(_FSMatrix):
    """
    Fixed Sparse Column format, in which the number of non-zero rows is fixed
    for each column. This format can also be used for block homothety matrices.

    """
    def __init__(self, shape, nrowmax=None, dtype=None, dtype_index=None,
                 data=None, verbose=False):
        _FSMatrix.__init__(self, 'fsc', shape, (1, 1), nrowmax, 0, dtype=dtype,
                           dtype_index=dtype_index, data=data, verbose=verbose)

    def _matvec(self, v, out=None):
        v, out, done = _FSMatrix._matvec(self, v, out, self.nrowmax)
        if done:
            return out

        data = self.data.reshape((-1, self.data.shape[-1]))
        extra_size = v.size // self.shape[1]
        v = v.reshape(-1, extra_size)
        out_ = out.reshape(-1, extra_size)
        if data.index.dtype.kind == 'u':
            for i in range(self.shape[1]):
                for b in data[i]:
                    out_[b.index] += b.value * v[i]
        else:
            for i in range(self.shape[1]):
                for b in data[i]:
                    if b.index >= 0:
                        out_[b.index] += b.value * v[i]
        return out

    def _transpose(self):
        return FSRMatrix(self.shape[::-1], data=self.data)


class FSRMatrix(_FSMatrix):
    """
    Fixed Sparse Row format, in which the number of non-zero columns is fixed
    for each row. This format can also be used for block homothety matrices.

    """
    def __init__(self, shape, ncolmax=None, dtype=None, dtype_index=None,
                 data=None, verbose=False):
        _FSMatrix.__init__(self, 'fsr', shape, (1, 1), ncolmax, 1, dtype=dtype,
                           dtype_index=dtype_index, data=data, verbose=verbose)

    def _matvec(self, v, out=None):
        v, out, done = _FSMatrix._matvec(self, v, out, self.ncolmax)
        if done:
            return out

        data = self.data.reshape((-1, self.data.shape[-1]))
        extra_size = v.size // self.shape[1]
        v = v.reshape(-1, extra_size)
        out_ = out.reshape(-1, extra_size)
        if data.index.dtype.kind == 'u':
            for i in range(self.shape[0]):
                b = data[i]
                out_[i] += np.sum(b.value[:, None] * v[b.index], axis=0)
        else:
            for i in range(self.shape[0]):
                b = data[i]
                b = b[b.index >= 0]
                out_[i] += np.sum(b.value[:, None] * v[b.index], axis=0)
        return out

    def _transpose(self):
        return FSCMatrix(self.shape[::-1], data=self.data)


class _FSBlockMatrix(_FSMatrix):
    def __init__(self, flib_id, shape, block_shape, nrowmax, sparse_axis,
                 dtype, dtype_index, data, verbose):
        if data is None:
            if block_shape is None:
                raise TypeError('The block shape is not specified.')
            block_shape = tointtuple(block_shape)
            if len(block_shape) != 2:
                raise ValueError(
                    "The number of dimensions of the blocks is not 2.")
        if block_shape == (1, 1):
            raise ValueError(
                'For block size of (1, 1), use the {}Matrix format.'
                .format(flib_id[:3].upper()))
        if any(_ not in (1, 2, 3) for _ in block_shape):
            raise NotImplementedError(
                "The sparse format {}BlockMatrix is not implemented for block"
                "s of shape '{}.'".format(flib_id[:3].upper(), block_shape))
        _FSMatrix.__init__(self, flib_id, shape, block_shape, nrowmax,
                           sparse_axis, dtype=dtype, dtype_index=dtype_index,
                           data=data, verbose=verbose)


class FSCBlockMatrix(_FSBlockMatrix):
    """
    Fixed Sparse Column Block format, in which the number of non-zero block
    rows is fixed for each block column.

    """
    def __init__(self, shape, block_shape=None, nrowmax=None, dtype=None,
                 dtype_index=None, data=None, verbose=False):
        if data is not None:
            block_shape = data.dtype['value'].shape[::-1]
        _FSBlockMatrix.__init__(self, 'fsc_block', shape, block_shape, nrowmax,
                                0, dtype, dtype_index, data, verbose)

    def _matvec(self, v, out=None):
        v, out, done = _FSMatrix._matvec(self, v, out, self.nrowmax)
        if done:
            return out

        data = self.data.reshape((-1, self.data.shape[-1]))
        out_ = out.reshape(-1, self.block_shape[0])
        v = v.reshape(-1, self.block_shape[1])
        for j in range(self.shape[1] // self.block_shape[1]):
            for b in data[j]:
                i = b['index']
                if i >= 0:
                    out_[i] += np.dot(b['value'].T, v[j])
        return out

    def _transpose(self):
        return FSRBlockMatrix(self.shape[::-1], data=self.data)

    def _get_dtype_data(self, dtype, dtype_index, block_shape):
        return [('index', dtype_index), ('value', dtype, block_shape[::-1])]


class FSRBlockMatrix(_FSBlockMatrix):
    """
    Fixed Sparse Row Block format, in which the number of non-zero block
    columns is fixed for each block row.

    """
    def __init__(self, shape, block_shape=None, ncolmax=None, dtype=None,
                 dtype_index=None, data=None, verbose=False):
        if data is not None:
            block_shape = data.dtype['value'].shape
        _FSMatrix.__init__(self, 'fsr_block', shape, block_shape, ncolmax, 1,
                           dtype, dtype_index, data, verbose)

    def _matvec(self, v, out=None):
        v, out, done = _FSMatrix._matvec(self, v, out, self.ncolmax)
        if done:
            return out

        data = self.data.reshape((-1, self.data.shape[-1]))
        out_ = out.reshape(-1, self.block_shape[0])
        v = v.reshape(-1, self.block_shape[1])
        for i in range(self.shape[0] // self.block_shape[0]):
            b = data[i]
            b = b[b.index >= 0]
            out_[i] += np.sum(np.einsum('...ij,...j->...i',
                                        b.value, v[b.index]), 0)
        return out

    def _transpose(self):
        return FSCBlockMatrix(self.shape[::-1], data=self.data)

    def _get_dtype_data(self, dtype, dtype_index, block_shape):
        return [('index', dtype_index), ('value', dtype, block_shape)]


class _FSRotation2dMatrix(_FSMatrix):
    def __init__(self, flib_id, shape, nmax, sparse_axis, dtype=None,
                 dtype_index=None, data=None, verbose=False):
        _FSMatrix.__init__(
            self, flib_id, shape, (2, 2), nmax, sparse_axis, dtype=dtype,
            dtype_index=dtype_index, data=data, verbose=verbose)

    def __mul__(self, other):
        if isscalarlike(other):
            data = self.data.copy()
            data.r11 *= other
            data.r21 *= other
            return type(self)(self.shape, data=data)
        other = np.asarray(other)
        if other.ndim == 1:
            return self._matvec(other)
        return NotImplemented

    def _get_dtype_data(self, dtype, dtype_index, block_shape):
        return [('index', dtype_index), ('r11', dtype), ('r21', dtype)]


class FSCRotation2dMatrix(_FSRotation2dMatrix):
    """
    Fixed Sparse Column format, in which the number of non-zero rows is fixed
    for each column. Each element of the sparse matrix is a 2x2 rotation
    matrix.

    """
    def __init__(self, shape,  nrowmax=None, dtype=None, dtype_index=None,
                 data=None, verbose=False):
        _FSRotation2dMatrix.__init__(
            self, 'fsc_rot2d', shape, nrowmax, 0, dtype=dtype,
            dtype_index=dtype_index, data=data, verbose=verbose)

    def _matvec(self, v, out=None):
        v, out, done = _FSMatrix._matvec(self, v, out, self.nrowmax)
        if done:
            return out

        data = self.data.reshape((-1, self.data.shape[-1]))
        out_ = out.reshape(-1, 2)
        v = v.reshape(-1, 2)
        for i in range(self.shape[1] // 2):
            for b in data[i]:
                if b.index >= 0:
                    out_[b.index, 0] +=  b.r11 * v[i, 0] + b.r21 * v[i, 1]
                    out_[b.index, 1] += -b.r21 * v[i, 0] + b.r11 * v[i, 1]
        return out

    def _transpose(self):
        return FSRRotation2dMatrix(self.shape[::-1], data=self.data)


class FSRRotation2dMatrix(_FSRotation2dMatrix):
    """
    Fixed Sparse Row format, in which the number of non-zero columns is fixed
    for each row. Each element of the sparse matrix is a 2x2 rotation matrix.

    """
    def __init__(self, shape, ncolmax=None, dtype=None, dtype_index=None,
                 data=None, verbose=False):
        _FSRotation2dMatrix.__init__(
            self, 'fsr_rot2d', shape, ncolmax, 1, dtype=dtype,
            dtype_index=dtype_index, data=data, verbose=verbose)

    def _matvec(self, v, out=None):
        v, out, done = _FSMatrix._matvec(self, v, out, self.ncolmax)
        if done:
            return out

        data = self.data.reshape((-1, self.data.shape[-1]))
        out_ = out.reshape(-1, 2)
        v = v.reshape(-1, 2)
        for i in range(self.shape[0] // 2):
            b = data[i]
            b = b[b.index >= 0]
            out_[i, 0] += np.sum(b.r11 * v[b.index, 0] -
                                 b.r21 * v[b.index, 1])
            out_[i, 1] += np.sum(b.r21 * v[b.index, 0] +
                                 b.r11 * v[b.index, 1])
        return out

    def _transpose(self):
        return FSCRotation2dMatrix(self.shape[::-1], data=self.data)


class _FSRotation3dMatrix(_FSMatrix):
    def __init__(self, flib_id, shape, nmax, sparse_axis, dtype=None,
                 dtype_index=None, data=None, verbose=False):
        _FSMatrix.__init__(
            self, flib_id, shape, (3, 3), nmax, sparse_axis, dtype=dtype,
            dtype_index=dtype_index, data=data, verbose=verbose)

    def __mul__(self, other):
        if isscalarlike(other):
            data = self.data.copy()
            data.r11 *= other
            data.r22 *= other
            data.r32 *= other
            return type(self)(self.shape, data=data)
        other = np.asarray(other)
        if other.ndim == 1:
            return self._matvec(other)
        return NotImplemented

    def _get_dtype_data(self, dtype, dtype_index, block_shape):
        return [('index', dtype_index),
                ('r11', dtype), ('r22', dtype), ('r32', dtype)]


class FSCRotation3dMatrix(_FSRotation3dMatrix):
    """
    Fixed Sparse Column format, in which the number of non-zero rows is fixed
    for each column. Each element of the sparse matrix is a 3x3 block
    whose transpose performs a rotation about the first component axis.

    """
    def __init__(self, shape, nrowmax=None, dtype=None, dtype_index=None,
                 data=None, verbose=False):
        _FSRotation3dMatrix.__init__(
            self, 'fsc_rot3d', shape, nrowmax, 0, dtype=dtype,
            dtype_index=dtype_index, data=data, verbose=verbose)

    def _matvec(self, v, out=None):
        v, out, done = _FSMatrix._matvec(self, v, out, self.nrowmax)
        if done:
            return out

        data = self.data.reshape(-1, self.nrowmax)
        out_ = out.reshape(-1, 3)
        v = v.reshape(-1, 3)
        for i in range(self.shape[1] // 3):
            for b in data[i]:
                if b.index >= 0:
                    out_[b.index, 0] +=  b.r11 * v[i, 0]
                    out_[b.index, 1] +=  b.r22 * v[i, 1] + b.r32 * v[i, 2]
                    out_[b.index, 2] += -b.r32 * v[i, 1] + b.r22 * v[i, 2]
        return out

    def _transpose(self):
        return FSRRotation3dMatrix(self.shape[::-1], data=self.data)


class FSRRotation3dMatrix(_FSRotation3dMatrix):
    """
    Fixed Sparse Row format, in which the number of non-zero columns is fixed
    for each row. Each element of the sparse matrix is a 3x3 block
    performing a rotation about the first component axis.

    """
    def __init__(self, shape, ncolmax=None, dtype=None, dtype_index=None,
                 data=None, verbose=False):
        _FSRotation3dMatrix.__init__(
            self, 'fsr_rot3d', shape, ncolmax, 1, dtype=dtype,
            dtype_index=dtype_index, data=data, verbose=verbose)

    def _matvec(self, v, out=None):
        v, out, done = _FSMatrix._matvec(self, v, out, self.ncolmax)
        if done:
            return out

        data = self.data.reshape((-1, self.data.shape[-1]))
        out_ = out.reshape(-1, 3)
        v = v.reshape(-1, 3)
        for i in range(self.shape[0] // 3):
            b = data[i]
            b = b[b.index >= 0]
            out_[i, 0] += np.sum(b.r11 * v[b.index, 0])
            out_[i, 1] += np.sum(b.r22 * v[b.index, 1] -
                                 b.r32 * v[b.index, 2])
            out_[i, 2] += np.sum(b.r32 * v[b.index, 1] +
                                 b.r22 * v[b.index, 2])
        return out_

    def _transpose(self):
        return FSCRotation3dMatrix(self.shape[::-1], data=self.data)


class SparseOperator(SparseBase):
    def __init__(self, arg, shapein=None, shapeout=None, dtype=None,
                 **keywords):
        if sp.issparse(arg):
            self.__class__ = pyoperators.linear.SparseOperator
            self.__init__(arg, dtype=None, **keywords)
            return
        if not isinstance(arg, (FSCMatrix, FSRMatrix,
                                FSCBlockMatrix, FSRBlockMatrix,
                                FSCRotation2dMatrix, FSRRotation2dMatrix,
                                FSCRotation3dMatrix, FSRRotation3dMatrix)):
            raise TypeError('The input sparse matrix type is not recognised.')
        if isinstance(arg, (FSCMatrix, FSRMatrix)):
            if shapein is None:
                bshapein = keywords.pop('broadcastable_shapein',
                                        (arg.shape[1],))
            else:
                shapein = tointtuple(shapein)
                test = np.cumprod(shapein) == arg.shape[1]
                if np.max(test) == 0:
                    bshapein = (arg.shape[1],)
                bshapein = shapein[:np.argmax(test)+1]
            self.broadcastable_shapein = bshapein
            if shapeout is None:
                bshapeout = keywords.pop('broadcastable_shapeout',
                                         (arg.shape[0],))
            else:
                shapeout = tointtuple(shapeout)
                test = np.cumprod(shapeout) == arg.shape[0]
                if np.max(test) == 0:
                    bshapeout = (arg.shape[0],)
                bshapeout = shapeout[:np.argmax(test)+1]
            self.broadcastable_shapeout = bshapeout
        else:
            bs = arg.block_shape
            if shapein is None:
                if bs[1] == 1:
                    shapein = arg.shape[1]
                else:
                    shapein = arg.shape[1] // bs[1], bs[1]
            if shapeout is None:
                if bs[0] == 1:
                    shapeout = arg.shape[0]
                else:
                    shapeout = arg.shape[0] // bs[0], bs[0]
        pyoperators.linear.SparseBase.__init__(
            self, arg, dtype=dtype, shapein=shapein, shapeout=shapeout,
            **keywords)
        self.set_rule('T', self._rule_transpose)
        self.set_rule(('T', '.'), self._rule_pTp, CompositionOperator)

    def direct(self, input, output, operation=operation_assignment):
        if operation is operation_assignment:
            output[...] = 0
        elif operation is not operator.iadd:
            raise ValueError(
                'Invalid reduction operation: {0}.'.format(operation))
        self.matrix._matvec(input, output)

    def todense(self, shapein=None, shapeout=None, inplace=False):
        shapein, shapeout = self._validate_shapes(shapein, shapeout)
        if shapein is None:
            raise ValueError("The operator's input shape is not explicit. Spec"
                             "ify it with the 'shapein' keyword.")
        extra_size = product(shapein) // self.matrix.shape[1]
        if self.shapein is None and extra_size > 1:
            shapein = shapein[:-1]
            broadcasting = True
        else:
            broadcasting = False
        out = SparseBase.todense(self, shapein=shapein, inplace=inplace)
        if broadcasting:
            out = np.kron(out, np.eye(extra_size))
        return out

    def reshapein(self, shape):
        if isinstance(self.matrix, (FSCMatrix, FSRMatrix)):
            return self.broadcastable_shapeout + \
                shape[len(self.broadcastable_shapein):]
        return self.shapeout

    def reshapeout(self, shape):
        if isinstance(self.matrix, (FSCMatrix, FSRMatrix)):
            return self.broadcastable_shapein + \
                shape[len(self.broadcastable_shapeout):]
        return self.shapein

    def validatein(self, shape):
        size = product(shape)
        if isinstance(self.matrix, (FSCMatrix, FSRMatrix)):
            if size % self.matrix.shape[1] != 0:
                raise ValueError(
                    "The input shape '{0}' is incompatible with the expected n"
                    "umber of column '{1}'.".format(shape,
                                                    self.matrix.shape[1]))
        self.matrix._validatein(shape)

    def validateout(self, shape):
        size = product(shape)
        if isinstance(self.matrix, (FSCMatrix, FSRMatrix)):
            if size % self.matrix.shape[0] != 0:
                raise ValueError(
                    "The output shape '{0}' is incompatible with the expected "
                    "number of rows '{1}'.".format(shape, self.matrix.shape[0]))
        self.matrix._validateout(shape)

    def corestrict(self, mask, inplace=False):
        """
        Corestrict the operator to a subspace defined by a mask
        (True means that the element is kept).

        """
        if not isinstance(self.matrix, (FSRMatrix, FSRRotation2dMatrix,
                                        FSRRotation3dMatrix)):
            raise NotImplementedError(
                'Corestriction is not implemented for {0} sparse storage.'
                .format(type(self.matrix).__name__))
        nrow = np.sum(mask)
        mask_ = np.repeat(mask, self.matrix.block_shape[0])
        out = self.copy()
        out.matrix = type(self.matrix)(
            (nrow * self.matrix.block_shape[0], self.matrix.shape[1]),
            self.matrix.data[mask_])
        if isinstance(self.matrix, FSRMatrix):
            out.broadcastable_shapeout = (nrow,)
            if self.shapeout is not None:
                ndims_out = len(self.broadcastable_shapeout)
                out.shapeout = (nrow,) + self.shapeout[ndims_out:]
        else:
            out.shapeout = (nrow, out.matrix.block_shape[0])
        if inplace:
            self.delete()
        return out

    def restrict(self, mask, inplace=False):
        """
        Restrict the operator to a subspace defined by a mask
        (True means that the element is kept). Indices are renumbered in-place
        if the inplace keyword is set to True.

        """
        if not isinstance(self.matrix, (FSRMatrix,
                                        FSRBlockMatrix,
                                        FSRRotation2dMatrix,
                                        FSRRotation3dMatrix)):
            raise NotImplementedError(
                'Restriction is not implemented for {0} sparse storage.'
                .format(type(self.matrix).__name__))
        flib_id = {FSRMatrix: '',
                   FSRBlockMatrix: '_block',
                   FSRRotation2dMatrix: '_rot2d',
                   FSRRotation3dMatrix: '_rot3d'}[type(self.matrix)]
        mask = np.asarray(mask)
        if mask.dtype != bool:
            raise TypeError('The mask is not boolean.')
        if isinstance(self.matrix, FSRMatrix):
            block_shapein = self.broadcastable_shapein
        else:
            block_shapein = self.shapein[:-1]
        if mask.shape != block_shapein:
            raise ValueError("Invalid shape '{}'. Expected value is '{}'.".
                             format(mask.shape, block_shapein))

        if inplace:
            matrix = self.matrix
        else:
            matrix = self.matrix.copy()
        itype = matrix.data.dtype['index']
        if itype.type in (np.int8, np.int16, np.int32, np.int64):
            f = 'fsr_restrict_i{0}'.format(itype.itemsize)
            func = getattr(fsp, f)
            block_shape = matrix.block_shape[0]
            ncol = func(matrix.data.view(np.int8).ravel(), mask.ravel(),
                        matrix.ncolmax, matrix.shape[0] // block_shape,
                        matrix.data.strides[-1])
        else:
            ncol = np.sum(mask)
            new_index = empty(mask.shape, itype)
            new_index[...] = -1
            new_index[mask] = np.arange(ncol, dtype=itype)
            undef = matrix.data.index < 0
            matrix.data.index = new_index[matrix.data.index]
            matrix.data.index[undef] = -1
        out = self.copy()
        matrix.shape = matrix.shape[0], ncol * matrix.block_shape[1]
        out.matrix = matrix
        if isinstance(self.matrix, FSRMatrix):
            out.broadcastable_shapein = (ncol,)
            if self.shapein is not None:
                ndims_in = len(self.broadcastable_shapein)
                out.shapein = (ncol,) + self.shapein[ndims_in:]
        else:
            out.shapein = (ncol, out.matrix.block_shape[1])
        if inplace:
            self.delete()
        return out

    def restrict_old(self, mask, inplace=False):
        """
        Restrict the operator to a subspace defined by a mask
        (True means that the element is kept). Indices are renumbered in-place
        if the inplace keyword is set to True.

        """
        if not isinstance(self.matrix, (FSRMatrix, FSRRotation2dMatrix,
                                        FSRRotation3dMatrix)):
            raise NotImplementedError(
                'Restriction is not implemented for {0} sparse storage.'
                .format(type(self.matrix).__name__))
        flib_id = {FSRMatrix: '',
                   FSRRotation2dMatrix: '_rot2d',
                   FSRRotation3dMatrix: '_rot3d'}[type(self.matrix)]
        mask = np.asarray(mask)
        if mask.dtype != bool:
            raise TypeError('The mask is not boolean.')
        if isinstance(self.matrix, FSRMatrix):
            block_shapein = self.broadcastable_shapein
        else:
            block_shapein = self.shapein[:-1]
        if mask.shape != block_shapein:
            raise ValueError("Invalid shape '{}'. Expected value is '{}'.".
                             format(mask.shape, block_shapein))

        if inplace:
            matrix = self.matrix
        else:
            matrix = self.matrix.copy()
        itype = matrix.data.index.dtype
        mtype = matrix.dtype
        if str(itype) in ('int32', 'int64') and \
           str(mtype) in ('float32', 'float64'):
            f = 'fsr{0}_restrict_i{1}_r{2}'.format(
                flib_id, itype.itemsize, mtype.itemsize)
            func = getattr(fop, f)
            block_shape = matrix.block_shape[0]
            ncol = func(matrix.data.view(np.int8).ravel(), mask.ravel(),
                        matrix.ncolmax, matrix.shape[0] // block_shape)
        else:
            ncol = np.sum(mask)
            new_index = empty(mask.shape, itype)
            new_index[...] = -1
            new_index[mask] = np.arange(ncol, dtype=itype)
            undef = matrix.data.index < 0
            matrix.data.index = new_index[matrix.data.index]
            matrix.data.index[undef] = -1
        out = self.copy()
        matrix.shape = matrix.shape[0], ncol * matrix.block_shape[1]
        out.matrix = matrix
        if isinstance(self.matrix, FSRMatrix):
            out.broadcastable_shapein = (ncol,)
            if self.shapein is not None:
                ndims_in = len(self.broadcastable_shapein)
                out.shapein = (ncol,) + self.shapein[ndims_in:]
        else:
            out.shapein = (ncol, out.matrix.block_shape[1])
        if inplace:
            self.delete()
        return out

    @staticmethod
    def _rule_transpose(self):
        if isinstance(self.matrix, (FSRMatrix, FSCMatrix)):
            keywords = {'broadcastable_shapein': self.broadcastable_shapeout,
                        'broadcastable_shapeout': self.broadcastable_shapein}
        else:
            keywords = {}
        return SparseOperator(self.matrix._transpose(), **keywords)

    @staticmethod
    def _rule_pTp(selfT, self):
        if not isinstance(
                self.matrix,
                (FSRMatrix, FSRRotation2dMatrix, FSRRotation3dMatrix)) or \
           self.matrix.ncolmax != 1 or self.shapein is None:
            return
        di = self.matrix.data.index.dtype
        dr = self.matrix.dtype
        dv = self.dtype
        if di not in (np.int32, np.int64) or \
           dr not in (np.float32, np.float64) or \
           dv not in (np.float32, np.float64):
            return
        f = 'fsc_{0}_ncolmax1_i{1}_r{2}_v{3}'.format(
            self.matrix._flib_id, di.itemsize, dr.itemsize, dv.itemsize)
        try:
            func = getattr(fsp, f)
        except AttributeError:
            return
        m = self.matrix.data.ravel().view(np.int8)
        data = np.zeros(self.shapein, dtype=self.dtype)
        func(m, data.ravel())
        return DiagonalOperator(data)


pyoperators.SparseOperator = SparseOperator



from __future__ import division

import healpy as hp
import healpy._healpy_pixel_lib as pixlib
import numpy as np
from pyoperators import CompositionOperator, IdentityOperator, Operator
from pyoperators.flags import inplace, real, square, symmetric
from pyoperators.utils import pi, strenum
from ...sparse import FSRMatrix, SparseOperator

__all__ = ['Healpix2CartesianOperator',
           'Cartesian2HealpixOperator',
           'Healpix2SphericalOperator',
           'Spherical2HealpixOperator',
           'HealpixConvolutionGaussianOperator',
           'HealpixLaplacianOperator']


@real
class _HealPixCartesian(Operator):
    def __init__(self, nside, nest=False, dtype=float, **keywords):
        self.nside = int(nside)
        self.nest = bool(nest)
        Operator.__init__(self, dtype=dtype, **keywords)

    @staticmethod
    def _reshapehealpix(shape):
        return shape + (3,)

    @staticmethod
    def _reshapecartesian(shape):
        return shape[:-1]

    @staticmethod
    def _validatecartesian(shape):
        if len(shape) == 0 or shape[-1] != 3:
            raise ValueError('Invalid cartesian shape.')

    @staticmethod
    def _rule_identity(o1, o2):
        if o1.nside == o2.nside and o1.nest == o2.nest:
            return IdentityOperator()


class Healpix2CartesianOperator(_HealPixCartesian):
    """
    Convert Healpix pixels into cartesian coordinates.

    """
    def __init__(self, nside, nest=False, **keywords):
        """
        nside : int
            Value of the map resolution parameter.
        nest : boolean, optional
            For the nested numbering scheme, set it to True. Default is
            ring scheme.

        """
        _HealPixCartesian.__init__(
            self, nside, nest=nest,
            reshapein=self._reshapehealpix,
            reshapeout=self._reshapecartesian,
            validateout=self._validatecartesian, **keywords)
        self.set_rule('I', lambda s:
                      Cartesian2HealpixOperator(s.nside, nest=s.nest))
        self.set_rule(('.', Cartesian2HealpixOperator), self._rule_identity,
                      CompositionOperator)

    def direct(self, input, output):
        input = input.astype(int)
        func = pixlib._pix2vec_nest if self.nest else pixlib._pix2vec_ring
        func(self.nside, input, output[..., 0], output[..., 1], output[..., 2])


class Cartesian2HealpixOperator(_HealPixCartesian):
    """
    Convert cartesian coordinates into Healpix pixels.

    """
    def __init__(self, nside, nest=False, **keywords):
        """
        nside : int
            Value of the map resolution parameter.
        nest : boolean, optional
            For the nested numbering scheme, set it to True. Default is
            ring scheme.

        """
        _HealPixCartesian.__init__(
            self, nside, nest=nest,
            reshapein=self._reshapecartesian,
            reshapeout=self._reshapehealpix,
            validatein=self._validatecartesian, **keywords)
        self.set_rule('I', lambda s:
                      Healpix2CartesianOperator(s.nside, nest=s.nest))
        self.set_rule(('.', Healpix2CartesianOperator), self._rule_identity,
                      CompositionOperator)

    def direct(self, input, output):
        func = pixlib._vec2pix_nest if self.nest else pixlib._vec2pix_ring
        func(self.nside, input[..., 0], input[..., 1], input[..., 2], output)


@real
class _HealPixSpherical(Operator):
    CONVENTIONS = ('zenith,azimuth',
                   'azimuth,zenith',
                   'elevation,azimuth',
                   'azimuth,elevation')

    def __init__(self, nside, convention, nest=False, dtype=float, **keywords):
        if not isinstance(convention, str):
            raise TypeError("The input convention '{0}' is not a string.".
                            format(convention))
        convention_ = convention.replace(' ', '').lower()
        if convention_ not in self.CONVENTIONS:
            raise ValueError(
                "Invalid spherical convention '{0}'. Expected values are {1}.".
                format(convention, strenum(self.CONVENTIONS)))
        self.nside = int(nside)
        self.convention = convention_
        self.nest = bool(nest)
        Operator.__init__(self, dtype=dtype, **keywords)

    @staticmethod
    def _reshapehealpix(shape):
        return shape + (2,)

    @staticmethod
    def _reshapespherical(shape):
        return shape[:-1]

    @staticmethod
    def _validatespherical(shape):
        if len(shape) == 0 or shape[-1] != 2:
            raise ValueError('Invalid spherical shape.')

    @staticmethod
    def _rule_identity(o1, o2):
        if o1.nside == o2.nside and o1.convention == o2.convention and \
           o1.nest == o2.nest:
            return IdentityOperator()


class Healpix2SphericalOperator(_HealPixSpherical):
    """
    Convert Healpix pixels into spherical coordinates in radians.


    The last dimension of the operator's output is 2 and it encodes
    the two spherical angles. Four conventions define what these angles are:
       - 'zenith,azimuth': (theta, phi) angles commonly used
       in physics or the (colatitude, longitude) angles used
       in the celestial and geographical coordinate systems
       - 'azimuth,zenith': (longitude, colatitude) convention
       - 'elevation,azimuth: (latitude, longitude) convention
       - 'azimuth,elevation': (longitude, latitude) convention

    """
    def __init__(self, nside, convention, nest=False, **keywords):
        """
        nside : int
            Value of the map resolution parameter.
        convention : string
            One of the following spherical coordinate conventions:
            'zenith,azimuth', 'azimuth,zenith', 'elevation,azimuth' and
            'azimuth,elevation'.
        nest : boolean, optional
            For the nested numbering scheme, set it to True. Default is
            ring scheme.

        """
        _HealPixSpherical.__init__(
            self, nside, convention, nest=nest,
            reshapein=self._reshapehealpix,
            reshapeout=self._reshapespherical,
            validateout=self._validatespherical, **keywords)
        self.set_rule(
            'I', lambda s: Spherical2HealpixOperator(s.nside, s.convention,
                                                     nest=s.nest))
        self.set_rule(('.', Spherical2HealpixOperator), self._rule_identity,
                      CompositionOperator)

    def direct(self, input, output):
        input = input.astype(int)
        func = pixlib._pix2ang_nest if self.nest else pixlib._pix2ang_ring
        if self.convention.startswith('azimuth'):
            o1, o2 = output[..., 1], output[..., 0]
        else:
            o1, o2 = output[..., 0], output[..., 1]
        func(self.nside, input, o1, o2)
        if 'elevation' in self.convention:
            np.subtract(0.5 * pi(self.dtype), o1, o1)


class Spherical2HealpixOperator(_HealPixSpherical):
    """
    Convert spherical coordinates in radians into Healpix pixels.

    The last dimension of the operator's output is 2 and it encodes
    the two spherical angles. Four conventions define what these angles are:
       - 'zenith,azimuth': (theta, phi) angles commonly used
       in physics or the (colatitude, longitude) angles used
       in the celestial and geographical coordinate systems
       - 'azimuth,zenith': (longitude, colatitude) convention
       - 'elevation,azimuth: (latitude, longitude) convention
       - 'azimuth,elevation': (longitude, latitude) convention

    """
    def __init__(self, nside, convention, nest=False, **keywords):
        """
        nside : int
            Value of the map resolution parameter.
        convention : string
            One of the following spherical coordinate conventions:
            'zenith,azimuth', 'azimuth,zenith', 'elevation,azimuth' and
            'azimuth,elevation'.
        nest : boolean, optional
            For the nested numbering scheme, set it to True. Default is
            ring scheme.

        """
        _HealPixSpherical.__init__(
            self, nside, convention, nest=nest,
            reshapein=self._reshapespherical,
            reshapeout=self._reshapehealpix,
            validatein=self._validatespherical, **keywords)
        self.set_rule(
            'I', lambda s: Healpix2SphericalOperator(s.nside, s.convention,
                                                     nest=s.nest))
        self.set_rule(('.', Healpix2SphericalOperator), self._rule_identity,
                      CompositionOperator)

    def direct(self, input, output):
        func = pixlib._ang2pix_nest if self.nest else pixlib._ang2pix_ring
        if self.convention.startswith('azimuth'):
            i1, i2 = input[..., 1], input[..., 0]
        else:
            i1, i2 = input[..., 0], input[..., 1]
        if 'elevation' in self.convention:
            i1 = 0.5 * pi(self.dtype) - i1
        func(self.nside, i1, i2, output)


@inplace
@real
@square
@symmetric
class HealpixConvolutionGaussianOperator(Operator):
    """
    Convolve a Healpix map by a gaussian kernel.

    """
    def __init__(self, fwhm=None, sigma=None, iter=3, lmax=None,
                 mmax=None, use_weights=False, datapath=None, dtype=float,
                 **keywords):
        """
        Keywords are passed to the Healpy function smoothing.

        """
        self.fwhm = fwhm
        self.sigma = sigma
        self.iter = iter
        self.lmax = lmax
        self.mmax = mmax
        self.use_weights = use_weights
        self.datapath = datapath
        Operator.__init__(self, dtype=dtype, **keywords)

    def direct(self, input, output):
        if input.ndim == 1:
            input = input[:, None]
            output = output[:, None]
        for i, o in zip(input.T, output.T):
            o[...] = hp.smoothing(
                i, fwhm=self.fwhm, sigma=self.sigma, iter=self.iter,
                lmax=self.lmax, mmax=self.mmax, use_weights=self.use_weights,
                datapath=self.datapath, verbose=False)

    def validatein(self, shape):
        if len(shape) == 0 or len(shape) > 2:
            raise ValueError('Invalid number of dimensions.')
        nside = int(np.round(np.sqrt(shape[0] / 12)))
        if 12 * nside**2 != shape[0]:
            raise ValueError(
                'The nside value cannot be inferred from the input number of p'
                "ixels '{0}'.".format(shape[0]))


@real
@square
@symmetric
class HealpixLaplacianOperator(SparseOperator):
    """
    9-Point stencil laplacian for Healpix maps.

    """
    def __init__(self, nside):
        """
        Parameters
        ----------
        nside : integer
            The Healpix map nside.

        """
        npix = 12 * nside**2
        ipix = np.arange(npix, dtype=np.int32)
        neighbours = hp.get_all_neighbours(nside, ipix)
        s = FSRMatrix((npix, npix), ncolmax=9, dtype_index=np.int32)
        s.data.index[:, 0] = ipix
        s.data.index[:, 1:] = neighbours.T
        h2 = 4 * np.pi / npix
        s.data.value[:, 0] = -20 / (6 * h2)
        s.data.value[:, 1:] = np.array([1, 4, 1, 4, 1, 4, 1, 4]) / (6 * h2)
        self.nside = nside
        SparseOperator.__init__(self, s)



from __future__ import division

import numpy as np
from numpy.testing import assert_allclose
from pyoperators.utils.testing import assert_eq, assert_raises
from pysimulators import (
    Acquisition, Instrument, MaskPolicy, PackedTable, Sampling, Scene)

flags = ['bad', 'u1', 'u2']


def test_mask_policy1():
    good_policy = ['kEep', 'removE', 'MASK']
    mask_policy = MaskPolicy(flags, good_policy)
    assert_eq(np.array(mask_policy), (0, 2, 1))
    assert mask_policy.bad == 'keep' and mask_policy.u1 == 'remove' and \
        mask_policy.u2 == 'mask'


def test_mask_policy2():
    bad_policy = ['remove', 'KKeep']
    assert_raises(ValueError, MaskPolicy, flags, bad_policy)
    assert_raises(KeyError, MaskPolicy, flags[0:2], bad_policy)


def test_get_noise():
    fsamp = 5
    sigma = 0.3
    scene = Scene(10)
    sampling = Sampling(2e4, period=1 / fsamp)
    shapes = ((1,), (2, 3))
    np.random.seed(0)

    class MyAcquisition1(Acquisition):
        def get_noise(self):
            return self.instrument.get_noise(self.sampling, sigma=sigma,
                                             fknee=fknee)

    def func1(shape, fknee):
        instrument = Instrument('', PackedTable(shape))
        acq = MyAcquisition1(instrument, sampling, scene)
        noise = acq.get_noise()
        assert noise.shape == (len(acq.instrument), len(acq.sampling))
        assert_allclose(np.std(noise), sigma, rtol=1e-2)
    for shape in shapes:
        for fknee in (0, 1e-7):
            yield func1, shape, fknee

    freq = np.arange(6) / 6 * fsamp
    psd = np.array([0, 1, 1, 1, 1, 1], float) * sigma**2 / fsamp

    class MyAcquisition2(Acquisition):
        def get_noise(self):
            return self.instrument.get_noise(self.sampling, psd=psd,
                                             bandwidth=freq[1], twosided=True)

    def func2(shape):
        instrument = Instrument('', PackedTable(shape))
        acq = MyAcquisition2(instrument, sampling, scene)
        noise = acq.get_noise()
        assert noise.shape == (len(acq.instrument), len(acq.sampling))
        assert_allclose(np.std(noise), sigma, rtol=1e-2)

    for shape in shapes:
        yield func2, shape

    freq = np.arange(4) / 6 * fsamp
    psd = np.array([0, 2, 2, 1], float) * sigma**2 / fsamp

    class MyAcquisition3(Acquisition):
        def get_noise(self):
            return self.instrument.get_noise(self.sampling, psd=psd,
                                             bandwidth=freq[1], twosided=False)

    def func3(shape):
        instrument = Instrument('', PackedTable(shape))
        acq = MyAcquisition3(instrument, sampling, scene)
        noise = acq.get_noise()
        assert noise.shape == (len(acq.instrument), len(acq.sampling))
        assert_allclose(np.std(noise), sigma, rtol=1e-2)

    for shape in shapes:
        yield func3, shape
